/*-------------------------------------------------
 *    Kernel version 2.4.20 - 2.6.8, 2.6.13
 *    RS485 (LPT-port)
 * 
 *    Andre Zaitsev, Olga Voziakova
 *    ovoz@sai.msu.ru
 *    $History:
 *
 *---------------------------------------- */
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kdev_t.h>
#include <linux/fs.h>

#include <linux/errno.h>
#include <linux/slab.h>
#include <asm/uaccess.h>   //  copy_from_user
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <linux/parport.h>

#include <mcua.h> // To define KERN_2_6
#ifdef KERN_2_6
   #include <linux/cdev.h>
   #include <linux/pnp.h>
#endif

#include <lpt.h>
#include <rs485.h>
#include <rs485_tools.h>
#include <rs485_irq.h>
#include <rserrno.h>
#include <rs485_fifo.h> 

static unsigned int major;
#ifdef KERN_2_6
static dev_t dev;
static struct cdev *cdev;
static unsigned int rs_nr_dev = 2;
static unsigned int minor = 0;
#endif

static int io1 = 0;
static int io2 = 0;

int irq1 = 0;
int irq2 = 0;

static int baudrate1 = 0;
static int baudrate2 = 0;

struct file_operations rs485_fops;
RS485INFO *info_port1, *info_port2;

static volatile int data_ready = 0;
DECLARE_WAIT_QUEUE_HEAD(tr_queue); // Test com-porta pokazal, chto eto medlennee prostogo oprosa (kak dlia 2.4)

#ifdef KERN_2_6
// S PnP svyazyvaemsia tol'ko dlya togo, chtoby activirovat' parport, inache 2.6.18 ne rabotaet..
// Sodrano s parport_pc.c
static const struct pnp_device_id rs_lpt_pnp_tbl[] = {
        /* Standard LPT Printer Port */
        {.id = "PNP0400", .driver_data = 0},
        /* ECP Printer Port */
        {.id = "PNP0401", .driver_data = 0},
        { }
};

MODULE_DEVICE_TABLE(pnp, rs_lpt_pnp_tbl);

static int rs_lpt_pnp_probe(struct pnp_dev *dev, const struct pnp_device_id *id)
{
    return 0;
}

static void rs_lpt_pnp_remove(struct pnp_dev *dev)
{ }


static struct pnp_driver rs_lpt_pnp_driver = {
    .name           = "parport_rs485",
    .id_table       = rs_lpt_pnp_tbl,
    .probe          = rs_lpt_pnp_probe,
    .remove         = rs_lpt_pnp_remove,
};

static int pnp_registered_parport;

void rs485_activate_parport(void) {
    int err;
    err = pnp_register_driver(&rs_lpt_pnp_driver);
    if(!err) pnp_registered_parport = 1;
}

void rs485_stop_parport(void) {
    if(pnp_registered_parport)
        pnp_unregister_driver(&rs_lpt_pnp_driver);
}
#endif

inline int rs485_getIO(int port) {
    if(port==2) return _IO_LPT2;
    return _IO_LPT1;
}

inline int rs485_getIRQ(int port) {
    if(port==2) return _IRQ_LPT2;
    return _IRQ_LPT1;
}

#ifndef KERN_2_6
static int rs_irq;
#endif
inline int rs485_get_port(int irq) {
#ifndef KERN_2_6
    rs_irq = irq;
#endif
    return (irq == irq1)? 1 : 2;
}

inline void rs485_allow_interrupt_card(RS485INFO *pinfo) {
    u8 a = inb(pinfo->base +EPP_CONTROL);
#ifdef __M_BARRIER__
    rmb();
#endif
    outb( a | EPP_INT_ENABLE, pinfo->base +EPP_CONTROL );
#ifdef __M_BARRIER__
    wmb();
#endif
}

inline void rs485_stop_interrupt_card(RS485INFO *pinfo) {
    u8 a = inb(pinfo->base +EPP_CONTROL);
#ifdef __M_BARRIER__
    rmb();
#endif
    outb( a & ~EPP_INT_ENABLE, pinfo->base +EPP_CONTROL );
#ifdef __M_BARRIER__
    wmb();
#endif
}

inline void rs485_transmit_disable(RS485INFO *pinfo) {
    u8 a = inb(pinfo->base +EPP_CONTROL);
#ifdef __M_BARRIER__
    rmb();
#endif
    outb( a | ENABLE_BI_DIRECT,
                pinfo->base +EPP_CONTROL );
#ifdef __M_BARRIER__
    wmb();
#endif
}

inline void rs485_transmit_enable(RS485INFO *pinfo) {
    u8 a = inb(pinfo->base +EPP_CONTROL);
#ifdef __M_BARRIER__
    rmb();
#endif
    outb( a & ~ENABLE_BI_DIRECT,
                        pinfo->base +EPP_CONTROL );
#ifdef __M_BARRIER__
    wmb();
#endif
}


int start_transmitter(RS485INFO *info) {

#ifndef KERN_2_6
    int ready=0;
    int resp_tmout=0;
#endif
    long remain_tm;
    const long timeout = HZ * RESP_TIMEOUT / 1000L / 1000L;    // JIFFIES

#ifdef __DEBUG_SIGNAL__
    outb(TXENABLE, 0x3f8 +UART_MCR);    // to COM1
#endif

    data_ready = 0;    // not ready
    rs485_transmit_enable(info);

    info->usCountByteOut = 0;

    while( 1 ) {

        info->usCountByteOut ++;
        if( info->usCountByteOut >= info->outbuffer.size ) {
            break;
        }

        outb(info->outbuffer.cache[info->usCountByteOut], // 1-, ,
                info->base +EPP_DATA);                        //  
#ifdef __M_BARRIER__
        wmb();
#endif
    }
    //   (     )
    outb( info->outbuffer.cache[0], info->base +EPP_ADDR );
#ifdef __M_BARRIER__
    wmb();
#endif
    info->status = PASSIVE_DATA;

    PDEBUGG( "send: %d bytes, QUIT\n",info->usCountByteOut );

#ifndef KERN_2_6
    rs485_transmit_disable(info);

//    ,    

    rs485_set_starttime();
    resp_tmout = 0;
    while( 1 ) {
//        rs485_stop_interrupts(); 
        disable_irq(rs_irq); // protect race condition with handler
        //   mem  ,       
//  mutex    ???
        ready = data_ready;
        enable_irq(rs_irq);
//        rs485_allow_interrupts();

        if(ready) break;
        if( rs485_timeout(RESP_TIMEOUT) ) {
            resp_tmout = 1;
            break;
        }
        udelay(RESP_WAIT);
    }
#ifdef __DEBUG_SIGNAL__
    outb(RXENABLE, 0x3f8 +UART_MCR);    // to COM1
#endif
    if(resp_tmout) {
        PDEBUG( "transmit: timeout!\n" );
        return -ERSNOCONVREPLY;
    } else {
        return(info->inbuffer_pas.size);
    }
#else
    remain_tm = wait_event_interruptible_timeout(tr_queue, data_ready != 0, timeout);
//       read_enable  ???  wait_event_interruptible_timeout ?
    rs485_transmit_disable(info);
#ifdef __DEBUG_SIGNAL__
    outb(RXENABLE, 0x3f8 +UART_MCR);    // to COM1
#endif
    if(signal_pending(current)) {
        return -ERESTARTSYS;
    }
    if( remain_tm == 0 ) {
        PDEBUG( "transmit: timeout!\n" );
        return -ERSNOCONVREPLY;
    }
    return(info->inbuffer_pas.size);    //    
#endif
}


void receiv_data(RS485INFO * info) {

/*
 * -----------------------------------------------------------
 *           .
 *          info->status == PASSIVE_DATA,    
 *         .   : [length][data...]  [signal]
 *          : [address][length][data...]
 * -----------------------------------------------------------
*/

    u8 iir = 0;
    int count_byte = 0;     //        (  )
    int ret = SUCCESS;
    struct _IN_BUFFER_ *current_inbuf;

#ifdef __DEBUG_TIMING__
    rs485_set_starttime();
#endif

#ifdef __DEBUG_SIGNAL__
    outb(TXENABLE, 0x3f8 +UART_MCR);    // to COM1
#endif

    // ,   -  
START:    
    iir = inb(info->base +EPP_STATUS);
#ifdef __M_BARRIER__
    rmb();
#endif
    if( (iir & PARPORT_STATUS_ACK) == 0 ) {
        goto QUIT_HANDLER;
    }

    info->signal = 0;
    info->usCountByteInp = 0; //   
    if( info->status == PASSIVE_DATA ) {
        current_inbuf = &(info->inbuffer_pas);
        current_inbuf->size = 0;    // yesli ne smozhem nichego polozhit'
        info->lSection = SECT_LENGTHBYTE;
    } else {
        current_inbuf = &(info->inbuffer_act);
        info->lSection = SECT_ADDRESSBYTE;
    }

    while( 1 ) {

        info->tmp_byte = inb( info->base +EPP_DATA );

        count_byte++;
        if(count_byte > CACHE_MAXSIZE) {
                PDEBUG("long handler\n");
                goto QUIT_HANDLER;
        }
                                                                                
        //  -  "-"
/*
        iir = inb(info->base +EPP_STATUS);
        mb();
        if( iir & EPP_TIMEOUT ) {

            PDEBUG( "receiver: timeout\n" );
            // reset:
            outb(EPP_TIMEOUT, info->base +EPP_STATUS);
            // wmb();
            if( (iir & PARPORT_STATUS_ACK) == 0 ) {
                PDEBUG( "receiver: timeout: No data in converter\n" );
                goto QUIT_HANDLER;    // no data in converter
            }
            info->tmp_byte = inb( info->base +EPP_DATA );
            mb();
            iir = inb(info->base +EPP_STATUS);
            mb();
            if( iir & EPP_TIMEOUT ) {
                PDEBUG( "receiver: repeat. timeout, quit\n" );
                goto QUIT_HANDLER;
            }
        }
*/

        switch(info->lSection) {

        case SECT_ADDRESSBYTE:

            //    addressbyte  cc !

            current_inbuf->cache[info->usCountByteInp] = info->tmp_byte;
            info->usCountByteInp ++;
            current_inbuf->adr  = info->tmp_byte;

            //      
            info->lSection = SECT_LENGTHBYTE;

            break;

        case SECT_LENGTHBYTE:

            //      ?
            if(info->tmp_byte >= PACKET_MAX_LENGTH) { //  
                if(info->tmp_byte < CACHE_MAXSIZE) { // length error
                    PDEBUG( "receiver: length error! lenght = %x\n", info->tmp_byte);
                } else {    // signal
                    if(info->status == PASSIVE_DATA ) {
                        current_inbuf->size = info->tmp_byte;
//  info->signal  ,      ,    
                        data_ready = 1;
                        wake_up(&tr_queue);
                        info->status = ACTIVE_DATA;
                    } else {    // Active data:
                        PDEBUG( "active receiver: signal = %x from address = %x received!\n",
                                     info->tmp_byte, current_inbuf->adr);
//  ne nuzhen on              info->signal = (char)info->tmp_byte;
//  vsio ravno pravilno       ret = rs485_put_data(info);
//  turbina ego ne prochtet   if(ret != SUCCESS) {
//                            PDEBUG("receiv_data: ret from rs485_put_data(SIGNAL) = %d\n", ret);
//                        }
                    }
                }
                goto START;
            }

            current_inbuf->cache[info->usCountByteInp] = info->tmp_byte;
            info->usCountByteInp ++;

            if( info->status == PASSIVE_DATA ) {
                current_inbuf->size = info->tmp_byte+1;
            } else {
                // +   
                current_inbuf->size = info->tmp_byte+2;
            }

            info->lSection = SECT_DATABYTES;

            break;

        case SECT_DATABYTES:            

            //  

            current_inbuf->cache[info->usCountByteInp] = info->tmp_byte;
            info->usCountByteInp ++;
            //       -   
            if( info->usCountByteInp < current_inbuf->size ) break;

            //  

            if( info->status == ACTIVE_DATA ) {
//                    printk("size=%d inp=%d\n",current_inbuf->size,info->usCountByteInp);
                //    FIFO
#ifdef __DEBUG_TIMING__
                if( rs485_timeout(100L) ) {    // 1 millisecond = 1000 usec
                    RS_PRINT( "Long packet: > 100 usec!\n" );
                }
#endif

                  rs485_put_data(info);
                  if(ret != SUCCESS) {
                      PDEBUG("receiv_data: ret from rs485_put_data(DATA) = %d\n", ret);
                  }

//          c... 
                goto START;    //    - ...
            } else {

                //      

                data_ready = 1;
                wake_up(&tr_queue);    // 2.4 ???
                info->status = ACTIVE_DATA;
                goto START;
            }

        }
    } //   

QUIT_HANDLER:

#ifdef __DEBUG_SIGNAL__
    outb(RXENABLE, 0x3f8 +UART_MCR); // to COM1
#endif

    return;
}

#define Fosc        14746
#define Minbaudrate 115
#define Maxbaudrate 1840

int ubrr_to_rate(int baudrate_code) {
    int a;
    a = baudrate_code + 1;
    if(a == 0) return 0;
    return( baudrate_code == 0x80 ? Fosc/8 : Fosc/16/a );
}

int rate_to_ubrr(int baudrate) {
    int baudrate_code;
    baudrate = baudrate < Minbaudrate ? Minbaudrate : baudrate;
    baudrate = baudrate > Maxbaudrate ? Maxbaudrate : baudrate;

    baudrate_code = baudrate > 1200 ? 0x80 : Fosc/16/baudrate-1;
//    baudrate_code = baudrate > 1200 ? 0x80 : iround(Fosc/16.0/baudrate)-1;
//VK       1800   -1 ( 0xFF ) 
//    
    return baudrate_code;
}

int rs485_set_baudrate(RS485INFO *pinfo) {
    int baudrate_code;
    baudrate_code = rate_to_ubrr(pinfo->baudrate);
    pinfo->baudrate = ubrr_to_rate(baudrate_code);
    PDEBUG("baudrate=%d\n",pinfo->baudrate);

    if( pinfo->interface_type == CONV_VERS1 ) {
//         udelay(10000);   
        pinfo->outbuffer.size = 9;
        pinfo->outbuffer.cache[pinfo->outbuffer.size] = 0;
        pinfo->outbuffer.cache[0] = 0x1f;
        pinfo->outbuffer.cache[1] = 0x07;
        pinfo->outbuffer.cache[2] = 0x1f;
        pinfo->outbuffer.cache[3] = 0x00;
        pinfo->outbuffer.cache[4] = baudrate_code;
        pinfo->outbuffer.cache[5] = 0x73;
        pinfo->outbuffer.cache[6] = 0x00;
        pinfo->outbuffer.cache[7] = 0x01;
        pinfo->outbuffer.cache[8] = 0x6c;
        start_transmitter(pinfo);
    }
    if( pinfo->interface_type == CONV_VERS2 ) {
        pinfo->outbuffer.size = 3;
        pinfo->outbuffer.cache[pinfo->outbuffer.size] = 0;
        pinfo->outbuffer.cache[0] = 0x1f;
        pinfo->outbuffer.cache[1] = 0x43;
        pinfo->outbuffer.cache[2] = baudrate_code;
          start_transmitter(pinfo);
    }
//    udelay(10000);
    return SUCCESS;
}

int rs485_test_device(RS485INFO *pinfo) {

    u8 iir = inb(pinfo->base+EPP_STATUS);
#ifdef __M_BARRIER__
    rmb();
#endif
    if( iir & PARPORT_STATUS_PAPEROUT ) {
        PDEBUG( "NO Converter!\n");
        return -ERSNOINTERFACE;
    }

    if( iir & PARPORT_STATUS_SELECT ) {
        PDEBUG( "NO POWER\n" );
        return -ERSSHUTDOWN;
    }

    if( iir & EPP_TIMEOUT ) {
        PDEBUG( "EPP TIMEOUT\n" );
        outb( EPP_TIMEOUT, pinfo->base +EPP_STATUS );
#ifdef __M_BARRIER__
        wmb();
#endif
        return -ERSISATMOUT;
    }
    return SUCCESS;
}

int rs485_reset(RS485INFO *pinfo) {
    u8 i;

    int ret = SUCCESS;
    ret = rs485_test_device(pinfo); // a to set_baudrate povesip komp!
    if(ret != SUCCESS) return ret;

    outb( EPP_TIMEOUT, pinfo->base +EPP_STATUS );        //  ?
#ifdef __M_BARRIER__
    wmb();
#endif
    outb( 0, pinfo->base +EPP_CONTROL );                //   !!!
#ifdef __M_BARRIER__
    wmb();
#endif
    udelay(EPP_WAIT);    // 250 mksec
    outb( EPP_RESET, pinfo->base +EPP_CONTROL );
#ifdef __M_BARRIER__
    wmb();
#endif
    udelay(EPP_WAIT);    // 250 mksec

    outb( 0x0e, pinfo->base +EPP_DATA );
#ifdef __M_BARRIER__
    wmb();
#endif
    outb( 0x1f, pinfo->base +EPP_ADDR );
#ifdef __M_BARRIER__
    wmb();
#endif
    udelay(RESP_WAIT);    // 100 mksec
// New code:
    i = inb(pinfo->base+EPP_DATA);
#ifdef __M_BARRIER__
    rmb();
#endif
      if( i == SignalACN ) {
          pinfo->interface_type = CONV_VERS2;
          outb( 0x87, pinfo->base +EPP_DATA ); //For new
#ifdef __M_BARRIER__
        wmb();
#endif
        outb( 0x1f, pinfo->base +EPP_ADDR );
#ifdef __M_BARRIER__
        wmb();
#endif
        udelay(RESP_WAIT);    // 100 mksec
        inb(pinfo->base+EPP_DATA);
#ifdef __M_BARRIER__
        rmb();
#endif
        PDEBUG("Converter code vers.2\n");
    } else  if (i == SignalACY) {
        pinfo->interface_type = CONV_VERS1;
        PDEBUG("Converter code vers.1\n");
    } else {
        PDEBUG( "reset: NO POWER\n" );
        return -ERSSHUTDOWN;    // na nekotorykh kompakh test_device ne srabatyvayet
                        
    }

    udelay(EPP_WAIT);

    pinfo->status = ACTIVE_DATA;
    rs485_transmit_disable(pinfo);
// Tut visnet komp naskvoz', esli ne vkliuchen pribor ili ne vkliuchen pribor ili net konvertera
// (Ne liuboj komp., no mnogie)

    rs485_allow_interrupt_card(pinfo);

    rs485_set_baudrate(pinfo);
    return ret;
}

// -------------------------------------------------------------------

void rs485_release_resource(void) {
    if( io2 != 0 ) release_region(io2, ADDRESS_REGION_SIZE);
    if( io1 != 0) release_region(io1, ADDRESS_REGION_SIZE);
    if( irq2 != 0 ) rs485_clean_irq((unsigned int) irq2);
    if( (irq1 != 0) && (irq1 != irq2) ) rs485_clean_irq((unsigned int) irq1);
    return;
}

int rs485_init(void) {        // IO ports and iterrupts initialization
    int err = 0;
#ifndef KERN_2_6
    if( (err=check_region(io1, ADDRESS_REGION_SIZE)) != 0 ) {
         RS_PRINT( KERN_CRIT "%s: I/O port 0x%08lX in use.\n",
                    ADAPTER_ID, (unsigned long) io1);
         io1 = 0;
         irq1 = 0;
         printk("io1=%x failed\n",io1);
    } else {
        request_region(io1, ADDRESS_REGION_SIZE, ADAPTER_ID);
    }
    if( (err=check_region(io2, ADDRESS_REGION_SIZE)) != 0 ) {
            RS_PRINT( KERN_CRIT "%s: I/O port 0x%08lX in use.\n",
                      ADAPTER_ID, (unsigned long) io2);
        io2 = 0;
        irq2 = 0;
        printk("io2=%x failed\n",io2);
    } else {
        request_region( io2, ADDRESS_REGION_SIZE, ADAPTER_ID );
    }    
#else // KERNEL 2.6
    if(!request_region(io1, ADDRESS_REGION_SIZE, ADAPTER_ID)) {
          RS_PRINT( KERN_CRIT "%s: I/O port 0x%08lX in use.\n",
                            ADAPTER_ID, (unsigned long) io1);
        io1 = 0;
        irq1 = 0;
        err = -EACCES;
    }
    if(!request_region(io2, ADDRESS_REGION_SIZE, ADAPTER_ID)) {
          RS_PRINT( KERN_CRIT "%s: I/O port 0x%08lX in use.\n",
                            ADAPTER_ID, (unsigned long) io2);
        io2 = 0;
        irq2 = 0;
        err = -EACCES;
    }    
#endif
    if( (io1 == 0) && (io2 == 0) ) return(err);

    // Interrupt init

    if(irq1 != 0) if( (err=rs485_request_irq(irq1)) != 0 ) {
        if( io1 != 0 ) release_region(io1, ADDRESS_REGION_SIZE);
        io1 = 0;
        irq1 = 0;
    }
    if((irq2 != 0) && (irq1 != irq2)) {
        if( (err=rs485_request_irq(irq2)) != 0 ) {
            if( io2 != 0 ) release_region(io2, ADDRESS_REGION_SIZE);
            io2 = 0;
            irq2 = 0;
        }
    }
    if( (irq1 == 0) && (irq2 == 0) ) return(err);
#ifndef KERN_2_6
    if( (major=register_chrdev(0, ADAPTER_ID, &rs485_fops)) <= 0) {
        RS_PRINT( KERN_CRIT "%s: unable to get major\n", ADAPTER_ID);
        rs485_release_resource();
        return major;
    } // else RS_PRINT(KERN_INFO "major=%d\n",major);

#else
    if( (err=alloc_chrdev_region(&dev, minor, rs_nr_dev, ADAPTER_ID)) < 0 ) {
        RS_PRINT( KERN_CRIT "%s: unable to get major\n", ADAPTER_ID);
        rs485_release_resource();
        return err;
    }
    major = MAJOR(dev);
//    RS_PRINT(KERN_INFO "major=%d\n",major);

    cdev = cdev_alloc();
    cdev->ops = &rs485_fops;
    cdev->owner = THIS_MODULE;
    err = cdev_add(cdev, dev, rs_nr_dev);
    if (err) {
        RS_PRINT(KERN_WARNING "rs485: device can't added to the system\n");
        return err;
    }
    rs485_activate_parport();
#endif
    return SUCCESS;
}

void rs485_blnumber_reset(RS485INFO *info) {
    int i;
    for( i=0; i<MAX_TASK_DEVICE; i++ ) {
        info->device[i].number = 0;
        info->device[i].round_count_prev = -1;
    }
    info->inbuffer_pas.adr = 0;
    info->inbuffer_act.adr = 0;
}

int rs485_open(struct inode *inode, struct file *pfile) {
    RS485INFO *info;
    int ret=0;
//    int imaj, imin;
    
#ifndef KERN_2_6
    switch(MINOR(inode->i_rdev)) {
#else
    switch(iminor(inode)) {
#endif
          case 0:
              if( io1 == 0 ) return -ERSBUSY;
              info = info_port1;
              break;

          case 1:
              if( io2 == 0) return -ERSBUSY;
              info = info_port2;
              break;

          default: return -ERSNOPORT;
    }

    pfile->private_data = info;

#ifndef KERN_2_6
    PDEBUG("%s: open device %d, %d\n", ADAPTER_ID, MAJOR(inode->i_rdev), MINOR(inode->i_rdev) );
#else
    PDEBUG("%s: open device %d, %d\n", ADAPTER_ID, imajor(inode), iminor(inode));
#endif

    if( info->use_mcua >= MAX_USE_STREAM ) return -ERSBUSY; // More than maximal shared users
    if( info->use_mcua == 0 ) {    // First opening
        rs485_fifo_init(info);
// test_device sidit v reset(), t.k. v com nado snachala delat'
// reset, a potom test_device, a v lpt naoborot.
        ret = rs485_reset(info); 
        if( ret < 0 ) return ret;
     }

    info->use_mcua ++;
#ifndef KERN_2_6
    MOD_INC_USE_COUNT;
#endif
    return SUCCESS;
}

int rs485_close(struct inode *inode, struct file *pfile) {

    RS485INFO *info = pfile->private_data;

//   if( down_interruptible(&info->sem ) ) {
//       up(&info->sem);
//       return -EINTR;
//   }

#ifndef KERN_2_6
     MOD_DEC_USE_COUNT;
#endif
     info->use_mcua --;

     if( info->use_mcua == 0 ) {
         rs485_stop_interrupt_card(info);
     }
//    udelay(SEM_WAIT);
//    up(&info->sem);

    return SUCCESS;
}

ssize_t rs485_write( struct file *fp, const char __user *buf, size_t size, loff_t *offset ) {

/*      ( NE   write )
 *    < 0  -  
 *  > 0x80      
 *  < 0x20        , (PACKET_MAX_LENGTH)
 */    
    RS485INFO    *info = fp->private_data;
    unsigned long rest = 0;
    int i, ret = 0;

    if( down_interruptible( &info->sem ) ) {
        up(&info->sem);
        return -EINTR;
    }
    if( (size == 0) || (size > PACKET_MAX_LENGTH ) ) {
        PDEBUG( "rs485_write: size=%d is inalid!\n", size);
        ret =  -ERSFORMAT;
        goto OUT;
    }
    if( (ret=rs485_test_device(info)) < 0 ) goto OUT;
    rest = copy_from_user( info->outbuffer.cache, buf, size );
    if(rest != 0) {
        PDEBUG("rs485_write: copy_from_user rest=%ld\n", rest);
        ret = -EFAULT;
        goto OUT;
    }
    info->outbuffer.size = size;
    for(i=0; i<MAX_HOP; i++) {
        ret = start_transmitter(info);
        if( (ret != SignalEER) && (ret != SignalNAK) ) break;
        msleep(4);
    }
    if( (ret > 0) && ( ret <= PACKET_MAX_LENGTH ) ) {
        rest = copy_to_user( (char*)buf, info->inbuffer_pas.cache, ret);
        if( rest != 0 ) {
            PDEBUG( "rs485_write: copy_to_user rest=%ld\n", rest );
            ret = -EFAULT;
            goto OUT;
        }       
    }
OUT:
    udelay(SEM_WAIT);
    up(&info->sem);
    return ret;
}

ssize_t rs485_read( struct file *fp, char __user *buf,
            size_t count, loff_t *offset ) {
/* 
 * Read data fifo
 */    
    char count0;
    RS485INFO *info = fp->private_data;
    int ret = 0;

    if( down_interruptible(&info->sem)) {        // 2.4 ???
        up(&info->sem);
        return -EINTR;
    }

    ret = rs485_data_fifo_read( info, buf, 1 );
    if( ret > 0 ) {
        get_user(count0, buf);    // buf in User space!!!
        if( count0 < (count-1) ) count = count0;
        else count--;
        ret = rs485_data_fifo_read( info, buf+1, count );
    }
//        udelay(SEM_WAIT);
    up(&info->sem);
    return ret;
}

int rs485_ioctl(struct inode *inode, struct file *fp,
            unsigned int cmd, unsigned long arg) {

    int err = SUCCESS;
    RS485INFO *rs485;
    unsigned long rest = 0;
    
    if(_IOC_TYPE(cmd) != RS_IOC_MAGIC) {        // 2.4 ???
        PDEBUG("ioctl_error!\n");
        return -ENOTTY;
    }
    if(_IOC_NR(cmd) > RS_IOC_MAXNR) {
        PDEBUG("ioctl_error!\n");
        return -ENOTTY;
    }
#ifdef KERN_2_6
/*
    if(_IOC_DIR(cmd) & _IOC_READ) {
        acc = access_ok(VERIFY_WRITE, (void __user *)arg, _IOC_SIZE(cmd));
    } else {
        acc = access_ok(VERIFY_READ, (void __user *)arg, _IOC_SIZE(cmd));
    }

    if(!acc) {
        PDEBUG("ioctl_error!\n");
        return -EFAULT;
    }
*/
#endif
    rs485 = fp->private_data;

    switch(cmd) {
        case RS_RESET:
            if( down_interruptible(&rs485->sem)) {
                up(&rs485->sem);
                return -EINTR;
            }
            err = rs485_reset(rs485);
            rs485_fifo_flush(rs485);
            udelay(SEM_WAIT);
            up(&rs485->sem);
            return err;

        case RS_FIFO_FLUSH:
            if( down_interruptible(&rs485->sem)) {
                up(&rs485->sem);
                return -EINTR;
            }
            rs485_fifo_flush(rs485);
            udelay(SEM_WAIT);
            up(&rs485->sem);
            return SUCCESS;

        case RS_SET_BAUDRATE:
            if( down_interruptible(&rs485->sem)) {
                up(&rs485->sem);
                return -EINTR;
            }
            err = rs485_test_device(rs485);
            if(err >= 0 ) {
                rs485->baudrate = (int)arg;
                err = rs485_set_baudrate(rs485);
            }
            udelay(SEM_WAIT);
            up(&rs485->sem);
            return err;

        case RS_GET_BAUDRATE:
            return rs485->baudrate;

        case RS_BLNUMBER_RESET:
            rs485_blnumber_reset(rs485);
            return SUCCESS;

        case RS_GET_VERS:
            rest = copy_to_user((char*)arg, RS485_UTS_RELEASE, sizeof(RS485_UTS_RELEASE));
            if(rest != 0) {
                PDEBUG("rs485_ioctl: get_vers: copy_to_user rest=%ld\n", rest);
                return -EFAULT;
            }
            return RS485_VERSION_CODE;

        case RS_GET_DATAFIFO_SIZE:
            return SIZE_STREAM_FIFO;

        default: break;
    }
    return -ERSNOIOCTL;
}

// ---------------------------------------------------------------------

struct file_operations rs485_fops = {
    .owner        = THIS_MODULE,
    .open         = rs485_open,
    .release      = rs485_close,
    .read         = rs485_read,
    .write        = rs485_write,
    .ioctl        = rs485_ioctl,
};


void rs485_free_mem1(void) {
    kfree(info_port1);
}

void rs485_free_mem2(void) {
    kfree(info_port2);
}

void rs485_delete_card(void) {
#ifdef KERN_2_6
    rs485_stop_parport();
#endif
    rs485_release_resource();
#ifndef KERN_2_6
    if ( (unregister_chrdev(major, ADAPTER_ID)) !=0){
        RS_PRINT(KERN_CRIT "%s: Error unload module\n", ADAPTER_ID);
    }
#else
    cdev_del(cdev);
    unregister_chrdev_region(dev,rs_nr_dev);
#endif
    return;
}

/* -------------------------------------------------
 *      
 *  :
 * io  <base address>
 * irq <interrupt>
 * --------------------------------------------------- */
 
MODULE_AUTHOR("Andre Zaitsev, Olga Voziakova (ovoz@sai.msu.ru)");
MODULE_DESCRIPTION("rs485 module");
#ifdef MODULE_LICENSE
    MODULE_LICENSE("GPL");
#endif

MODULE_PARM_DESC(io1, "Base LPT1-port (hex), default io1=_IO_LPT1");
MODULE_PARM_DESC(io2, "Base LPT2-port (hex), default io2=_IO_LPT2");
MODULE_PARM_DESC(irq1, "Interrupt LPT1-port,default irq1=_IRQ_LPT1");
MODULE_PARM_DESC(irq2, "Interrupt LPT2-port,default irq2=_IRQ_LPT2");

MODULE_PARM_DESC(io1, "i");
MODULE_PARM_DESC(io2, "i");
MODULE_PARM_DESC(irq1, "i");
MODULE_PARM_DESC (irq2, "i");

//MODULE_PARM_DESC(portnumb, "PORT NUMBER, default portnumb=_PORT_NUMB");
//MODULE_PARM(portnumb, "i");

MODULE_PARM_DESC(baudrate1, "BAUD RATE UART port1, default baudrate=_BAUDRATE_1");
MODULE_PARM_DESC(baudrate2, "BAUD RATE UART port2, default baudrate=_BAUDRATE_2");
MODULE_PARM_DESC(baudrate1, "i");
MODULE_PARM_DESC(baudrate2, "i");

#ifndef KERN_2_6
EXPORT_NO_SYMBOLS;
#endif

int init_module(void) {
    int err = 0;
    int err1 = 0;
    int err2 = 0;
    char version[]=RS485_UTS_RELEASE;

//    portnumb = portnumb ? portnumb : _PORT_NUMB;    //     insmod...
    baudrate1 = baudrate1 ? baudrate1 : _BAUDRATE_1;
    baudrate2 = baudrate2 ? baudrate2 : _BAUDRATE_2;

    PDEBUG("baudrate1=%d baudrate2=%d\n", baudrate1, baudrate2);

    io1 = io1 ? io1 : rs485_getIO(1);
    io2 = io2 ? io2 : rs485_getIO(2);
    irq1 = irq1 ? irq1 : rs485_getIRQ(1);
    irq2 = irq2 ? irq2 : rs485_getIRQ(2);

    info_port1 = (RS485INFO *) kmalloc (sizeof (RS485INFO), GFP_KERNEL);
    if(info_port1 == NULL){
        err = -ENOMEM;
        goto FAIL_1;
    }
    
    info_port2 = (RS485INFO *) kmalloc (sizeof (RS485INFO), GFP_KERNEL);
    if(info_port2 == NULL){
        err = -ENOMEM;
        goto FAIL_2;
    }
    
    memset((void *)info_port1, 0 , sizeof(RS485INFO));
    memset((void *)info_port2, 0 , sizeof(RS485INFO));

    rs485_blnumber_reset(info_port1);
    rs485_blnumber_reset(info_port2);

    if( (err=rs485_init()) != 0 ) goto FAIL_3;

    info_port1->base = io1;
    info_port2->base = io2;

    info_port1->baudrate = baudrate1;
    info_port2->baudrate = baudrate2;

    rs485_stop_interrupt_card(info_port1);
    rs485_stop_interrupt_card(info_port2);

    init_MUTEX(&info_port1->sem);        // 2.4 ???
    init_MUTEX(&info_port2->sem);

//    

    if( ((err1=rs485_fifo_create(info_port1)) != 0)
        || ((err2=rs485_fifo_create(info_port2)) != 0) ) {
        err = err1 ? err1 : err2;
        goto FAIL_5;
    }

    RS_PRINT(KERN_INFO "%s module, version %s\n",
                ADAPTER_ID, version);
    RS_PRINT(KERN_INFO "%s: install card on the base=0x%X, 0x%X and irq=%d, %d\n",
                ADAPTER_ID, io1, io2, irq1, irq2);

    return SUCCESS;

    rs485_fifo_destroy();
FAIL_5:    //    create_fifo failed
    rs485_delete_card();
FAIL_3:
    rs485_free_mem2();
FAIL_2:
    rs485_free_mem1();
FAIL_1:
    return err;
}

void cleanup_module(void){
    rs485_fifo_destroy();
    rs485_free_mem2();
    rs485_free_mem1();
    rs485_delete_card();
    return;
}
